# Python Imports
from itertools import chain

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtWidgets import QWidget

# Third Party Imports
import numpy as np

# Own Imports
from kmap import __directory__
from kmap.library.qwidgetsub import Tab
from kmap.controller.sliceddatatab import SlicedDataTab
from kmap.controller.orbitaldatatab import OrbitalDataTab
from kmap.controller.matplotlibwindow import MatplotlibLineWindow

# Load .ui File
UI_file = __directory__ / 'ui/profileplottab.ui'
ProfilePlotTab_UI, _ = uic.loadUiType(UI_file)


class ProfilePlotTab(Tab, ProfilePlotTab_UI):

    def __init__(self, tab_widget, title, *args, **kwargs):
        # Setup GUI
        super(ProfilePlotTab, self).__init__(*args, **kwargs)
        self.setupUi(self)

        # List of SlicedDataTabs and OrbitalDataTabs this ProfilePlotTab
        # know about
        self.tabs = []
        # For each tab in self.tabs a 6-element long sublist with bools
        # denoted which plots options (checkboxes) are enabled for the
        # corresponding tab
        self.show_options = []
        self.title = title

        self._setup()
        self._connect(tab_widget)

        self.load_tabs(tab_widget)

    def refresh_plot(self):
        is_slice_plot = self.slice_radiobutton.isChecked()
        is_line_plot = self.line_radiobutton.isChecked()
        phi_sample = self.phi_sample_spinbox.value()
        line_sample = self.line_sample_spinbox.value()
        normalized = self.normalize_checkbox.isChecked()
        x_label = 'None'
        self.plot_item.clear()

        if is_slice_plot:
            regions = np.array(['center', 'x', 'y', 'roi', 'border',
                                'ring'])[
                self.show_options[self.tab_combobox.currentIndex()]]

            if self.center_checkbox.isChecked():
                np.insert(regions, 0, 'center')

            if not list(regions):
                return

            for region in regions:
                tab = self.current_tab()
                axis = tab.get_axis()
                data = [tab.get_data(), axis]
                crosshair = tab.get_crosshair().model
                title = tab.get_title()
                self.plot_item.plot(data, title, crosshair,
                                    region=region,
                                    phi_sample=phi_sample,
                                    line_sample=line_sample,
                                    normalized=normalized)
                axis = tab.get_data().axes[axis]
                x_label = '%s [%s]' % (axis.label, axis.units)

        else:
            for tab, show_options in zip(self.tabs, self.show_options):
                data = tab.get_displayed_plot_data()
                bottom_label, left_label = tab.get_plot_labels()
                crosshair = tab.get_crosshair().model
                title = tab.get_title()
                x_label = ''

                if is_line_plot and show_options[1]:
                    self.plot_item.plot(data, title, crosshair,
                                        region='x',
                                        phi_sample=phi_sample,
                                        line_sample=line_sample,
                                        normalized=normalized)
                    x_label = left_label

                if is_line_plot and show_options[2]:
                    self.plot_item.plot(data, title, crosshair,
                                        region='y',
                                        phi_sample=phi_sample,
                                        line_sample=line_sample,
                                        normalized=normalized)
                    x_label = bottom_label

                if not is_line_plot and show_options[3]:
                    self.plot_item.plot(data, title, crosshair,
                                        region='roi',
                                        phi_sample=phi_sample,
                                        line_sample=line_sample,
                                        normalized=normalized)
                    x_label = 'Angle Phi [°]'

                if not is_line_plot and show_options[4]:
                    self.plot_item.plot(data, title, crosshair,
                                        region='border',
                                        phi_sample=phi_sample,
                                        line_sample=line_sample,
                                        normalized=normalized)
                    x_label = 'Angle Phi [°]'

                if not is_line_plot and show_options[5]:
                    self.plot_item.plot(data, title, crosshair,
                                        region='ring',
                                        phi_sample=phi_sample,
                                        line_sample=line_sample,
                                        normalized=normalized)
                    x_label = 'Angle Phi [°]'

        self.plot_item.set_label(x_label, 'Intensity [a.u.]')

    def save_state(self):
        if self.circle_radiobutton.isChecked():
            type_ = 'circle'

        elif self.line_radiobutton.isChecked():
            type_ = 'line'

        else:
            type_ = 'slice'

        states = []
        checkboxes = [self.center_checkbox, self.x_checkbox,
                      self.y_checkbox, self.roi_checkbox,
                      self.border_checkbox, self.annulus_checkbox]
        for checkbox in checkboxes:
            states.append(checkbox.checkState())

        sample_rate = [self.phi_sample_spinbox.value(),
                       self.line_sample_spinbox.value()]

        save = {'title': self.get_title(),
                'type': type_,
                'tab_index': self.tab_combobox.currentIndex(),
                'states': states,
                'sample_rate': sample_rate}

        return save, []

    def restore_state(self, save):
        type_ = save['type']

        if type_ == 'circle':
            self.circle_radiobutton.click()

        elif type_ == 'line':
            self.line_radiobutton.click()

        elif type_ == 'slice':
            self.slice_radiobutton.click()

        else:
            raise ValueError

        tab_index = save['tab_index']
        self.tab_combobox.setCurrentIndex(tab_index)

        sample_rate = save['sample_rate']
        self.phi_sample_spinbox.setValue(sample_rate[0])
        self.line_sample_spinbox.setValue(sample_rate[1])

        checkboxes = [self.center_checkbox, self.x_checkbox,
                      self.y_checkbox, self.roi_checkbox,
                      self.border_checkbox, self.annulus_checkbox]
        for state, checkbox in zip(save['states'], checkboxes):
            checkbox.setCheckState(state)

        self.refresh_button.click()

    def load_tabs(self, tab_widget):
        for tab in tab_widget.get_tabs_of_type(None):
            self.add_tab(tab)

    def add_tab(self, tab):
        if type(tab) in [SlicedDataTab, OrbitalDataTab]:
            tab.close_requested.connect(self._remove_tab)
            self.tab_combobox.blockSignals(True)
            self.tab_combobox.addItem(tab.get_title())
            self.tab_combobox.blockSignals(False)
            self.tabs.append(tab)
            self.show_options.append(6 * [False])

    def export_to_txt(self):
        data = self.plot_item.get_data()

        text = ''

        for data_set in data:
            name = data_set['name']
            x = data_set['x']
            y = data_set['y']
            text += '# ' + name + '\n'
            for xi, yi in zip(x, y):
                text += '%g  %g \n' % (xi, yi)
            text += '\n'

        return text

    def display_in_matplotlib(self):
        data = self.plot_item.get_data()

        window = MatplotlibLineWindow(data)

        return window

    def current_tab(self):
        return self.tabs[self.tab_combobox.currentIndex()]

    def _remove_tab(self):
        tab = self.sender()

        if tab in self.tabs:
            index = self.tabs.index(tab)
            self.tab_combobox.removeItem(index)
            del self.tabs[index]
            del self.show_options[index]

            self.refresh_plot()

    def rename_loaded_tab(self, tab, new_name):
        try:
            index = self.tabs.index(tab)
            self.tab_combobox.setItemText(index, new_name)

        except ValueError:
            pass

    def switch_type(self):
        # Widgets associated with line-like plotting (e.g. x/y - Line)
        line_widgets = [self.x_checkbox, self.y_checkbox]
        # Widgets associated with ring-like plotting (e.g. ROI, Annulus)
        circle_widgets = [self.annulus_checkbox,
                          self.border_checkbox, self.roi_checkbox]

        show_slice_bool = self.slice_radiobutton.isChecked()
        show_line_bool = True if self.line_radiobutton.isChecked(
        ) or show_slice_bool else False
        show_circle_bool = True if self.circle_radiobutton.isChecked(
        ) or show_slice_bool else False

        for widget in chain(line_widgets, circle_widgets):
            widget.blockSignals(True)

            if widget in line_widgets:
                widget.setVisible(show_line_bool)

            if widget in circle_widgets:
                widget.setVisible(show_circle_bool)

            widget.blockSignals(False)

        self.center_checkbox.setVisible(show_slice_bool)

        other_widgets = [self.phi_sample_label, self.phi_sample_spinbox,
                         self.line_sample_label, self.line_sample_spinbox]

        for widget in other_widgets:
            widget.setVisible(not show_slice_bool)

    def _change_tab(self, index):
        if isinstance(self.tabs[index], SlicedDataTab):
            self.slice_radiobutton.setVisible(True)

        else:
            self.slice_radiobutton.setVisible(False)

            if self.slice_radiobutton.isChecked():
                self.line_radiobutton.setChecked(True)

        checkboxes = [self.center_checkbox,
                      self.x_checkbox, self.y_checkbox,
                      self.roi_checkbox, self.border_checkbox,
                      self.annulus_checkbox]

        for checkbox, show in zip(checkboxes, self.show_options[index]):
            checkbox.blockSignals(True)
            checkbox.setChecked(show)
            checkbox.blockSignals(False)

    def _change_checkbox(self):
        checkbox = self.sender()
        checkboxes = [self.center_checkbox,
                      self.x_checkbox, self.y_checkbox,
                      self.roi_checkbox, self.border_checkbox,
                      self.annulus_checkbox]

        tab_index = self.tab_combobox.currentIndex()
        if tab_index != -1:
            checkbox_index = checkboxes.index(checkbox)
            self.show_options[tab_index][checkbox_index] = checkbox.isChecked()

    def _setup(self):
        self.switch_type()

    def _connect(self, tab_widget):
        tab_widget.tab_added.connect(self.add_tab)

        self.refresh_button.clicked.connect(self.refresh_plot)
        self.circle_radiobutton.toggled.connect(self.switch_type)
        self.slice_radiobutton.toggled.connect(self.switch_type)

        self.tab_combobox.currentIndexChanged.connect(self._change_tab)
        self.center_checkbox.stateChanged.connect(self._change_checkbox)
        self.x_checkbox.stateChanged.connect(self._change_checkbox)
        self.y_checkbox.stateChanged.connect(self._change_checkbox)
        self.roi_checkbox.stateChanged.connect(self._change_checkbox)
        self.border_checkbox.stateChanged.connect(self._change_checkbox)
        self.annulus_checkbox.stateChanged.connect(self._change_checkbox)



# Python Imports
import logging

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtWidgets import QVBoxLayout

# Own Imports
from kmap import __directory__
from kmap.library.qwidgetsub import Tab
from kmap.model.splitview_model import SplitViewTabModel
from kmap.controller.dataslider import DataSliderNoTranspose
from kmap.controller.colormap import Colormap
from kmap.controller.crosshairannulus import CrosshairAnnulus
from kmap.controller.interpolation import Interpolation
from kmap.controller.splitviewoptions import SplitViewOptions
from kmap.controller.matplotlibwindow import MatplotlibImageWindow
from kmap.config.config import config

# Load .ui File
UI_file = __directory__ / 'ui/splitviewtab.ui'
SplitViewTab_UI, _ = uic.loadUiType(UI_file)


class SplitViewTab(Tab, SplitViewTab_UI):

    def __init__(self, sliced_tab, orbital_tab):
        self.sliced_tab = sliced_tab
        self.orbital_tab = orbital_tab

        # Setup GUI
        super(SplitViewTab, self).__init__()
        self.setupUi(self)

        self._setup()
        self._connect()

        self.change_axis(0)
        self.change_slice(0)

    @classmethod
    def init_from_save(cls, save, sliced_tab, orbital_tab):
        tab = SplitViewTab(sliced_tab, orbital_tab)

        tab.slider.restore_state(save['slider']),
        tab.crosshair.restore_state(save['crosshair']),
        tab.interpolation.restore_state(save['interpolation']),
        tab.split_options.restore_state(save['split_options'])

        return tab

    def save_state(self):
        save = {'title': self.title,
                'slider': self.slider.save_state(),
                'crosshair': self.crosshair.save_state(),
                'interpolation': self.interpolation.save_state(),
                'split_options': self.split_options.save_state()}

        return save, [self.sliced_tab, self.orbital_tab]

    def get_title(self):
        return 'Split View'

    def get_axis(self):
        return self.slider.get_axis()

    def get_slice(self):
        return self.slider.get_index()

    def change_slice(self, index=-1):
        axis = self.slider.get_axis()
        slice_index = index if index != -1 else self.get_slice()
        data = self.model.update_displayed_plot_data(slice_index, axis)

        self.plot_item.plot(data)
        self.crosshair.update_label()

    def change_axis(self, axis):
        # 'axes' is a copy of all axes except the one with index 'axis'
        axes = [a for i, a in enumerate(self.model.get_sliced_data().axes) if
                i != axis]
        index = self.get_slice()
        data = self.model.update_displayed_plot_data(index, axis)

        self.plot_item.set_labels(axes[1], axes[0])
        self.interpolation.set_label(axes[1], axes[0])
        self.plot_item.plot(data)

    def crosshair_changed(self):
        self.crosshair.update_label()

    def display_in_matplotlib(self):
        data = self.model.displayed_plot_data
        LUT = self.plot_item.get_LUT()

        window = MatplotlibImageWindow(data, LUT=LUT)

        return window

    def change_split_options(self, scale, type_):
        self.model.set_scale(scale)
        self.model.set_type(type_)
        self.change_slice()

    def change_symmetry(self, symmetry, mirror):
        self.model.change_symmetry(symmetry, mirror)
        self.change_axis(self.slider.get_axis())

    def closeEvent(self, event):
        del self.model

        Tab.closeEvent(self, event)

    def _setup(self):
        self.crosshair = CrosshairAnnulus(self.plot_item)
        self.colormap = Colormap(self.plot_item)
        self.interpolation = Interpolation()
        self.interpolation.set_force_interpolation(True)
        self.model = SplitViewTabModel(self.sliced_tab, self.orbital_tab,
                                       self.interpolation)
        self.split_options = SplitViewOptions()

        self.slider = DataSliderNoTranspose(self.model.get_sliced_data())

        layout = self.scroll_area.widget().layout()
        layout.insertWidget(0, self.slider)
        layout.insertWidget(1, self.split_options)
        layout.insertWidget(2, self.interpolation)
        layout.insertWidget(3, self.colormap)
        layout.insertWidget(4, self.crosshair)

    def _connect(self):
        self.crosshair.crosshair_changed.connect(self.crosshair_changed)
        self.slider.slice_changed.connect(self.change_slice)
        self.slider.axis_changed.connect(self.change_axis)
        self.slider.symmetry_changed.connect(self.change_symmetry)
        self.interpolation.smoothing_changed.connect(self.change_slice)
        self.interpolation.interpolation_changed.connect(self.change_slice)
        self.split_options.values_changed.connect(self.change_split_options)



from setuptools import setup, find_packages
from kmap import __version__

setup(name='kMap',
      version=__version__,
      description='Display, modify and analyse ARPES experiments and DFT \
      simulations.',
      packages=find_packages(),
      license='MIT',
      url='https://github.com/brands-d/kMap',
      author='Dominik Brandstetter',
      author_email='dominik.brandstetter@edu.uni-graz.at',
      zip_safe=False,
      install_requires=['h5py>=3.1.0', 'scipy>=1.5.1',
                        'matplotlib>=3.3.0', 'numpy>=1.19.0,!=1.19.4',
                        'PyQt5>=5.15.0', 'PyOpenGL>=3.1.5',
                        'lmfit>=1.0.1', 'pyqtgraph>=0.11.0',
                        'qdarkstyle>=2.8.1'],
      include_package_data=True)



# Python Imports
import logging

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QWidget

# Own Imports
from kmap import __directory__

# Load .ui File
UI_file = __directory__ / 'ui/splitviewoptions.ui'
SplitViewOptions_UI, _ = uic.loadUiType(UI_file)


class SplitViewOptions(QWidget, SplitViewOptions_UI):
    values_changed = pyqtSignal(float, str)

    def __init__(self, *args, **kwargs):
        # Setup GUI
        super(SplitViewOptions, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._connect()

    def save_state(self):
        scale = self.scale_spinbox.value()
        type_ = self.type_combobox.currentIndex()

        save = {'Scale': scale, 'Type': type_}

        return save

    def restore_state(self, save):
        scale = save['Scale']
        type_ = save['Type']

        self.scale_spinbox.setValue(scale)
        self.type_combobox.setCurrentIndex(type_)

    def get_parameters(self):
        scale = self.scale_spinbox.value()
        type_ = self.type_combobox.currentText()

        return scale, type_

    def emit_values(self):
        scale, type_ = self.get_parameters()
        self.values_changed.emit(scale, type_)

    def _connect(self):
        self.scale_spinbox.valueChanged.connect(self.emit_values)
        self.type_combobox.currentIndexChanged.connect(self.emit_values)



# Python Imports
import _pickle as pickle
import bz2
import logging

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtGui import QIcon, QKeySequence
from PyQt5.QtWidgets import QMainWindow, QFileDialog, QMessageBox

# Own Imports
from kmap import __directory__
from kmap.config.config import config
from kmap.controller.databasewindows import (
    OrbitalDatabase, SlicedDatabaseWindow)
from kmap.controller.orbitaldatatab import OrbitalDataTab
from kmap.controller.sliceddatatab import SlicedDataTab
from kmap.controller.tabchoosewindow import TabChooseWindow
from kmap.model.mainwindow_model import MainWindowModel

# Load .ui File
UI_file = __directory__ / 'ui/mainwindow.ui'
MainWindow_UI, _ = uic.loadUiType(UI_file)


class MainWindow(QMainWindow, MainWindow_UI):

    def __init__(self):
        # Setup GUI
        super(MainWindow, self).__init__()
        self.setupUi(self)
        self._setup()
        self._connect()

        self.sub_windows = {}

        self.model = MainWindowModel(self)

        self.open_welcome()

    def load_hdf5_files(self):
        # Load one or more new hdf5 files

        log = logging.getLogger('kmap')
        log.info('Loading .hdf5 file(s)...')

        extensions = 'hdf5 files (*.hdf5 *.h5);; All Files (*)'

        path = config.get_key('paths', 'hdf5_start')
        if path == 'None':
            paths, _ = QFileDialog.getOpenFileNames(
                None, 'Open file(s)', filter=extensions)
        else:
            start_path = str(__directory__ / path)
            paths, _ = QFileDialog.getOpenFileNames(
                None, 'Open file(s)', str(start_path), extensions)

        if not paths:
            # No file chosen
            log.info('No file chosen')
            return

        for path in paths:
            self.tab_widget.open_sliced_data_tab(path)

    def open_orbitaldatabase_browser(self):
        database = OrbitalDatabase()
        database.files_chosen.connect(self.load_cube_files_online)

        self.sub_windows.update({str(id(database)): database})

    def open_binding_energy_sliceddatabase_browser(self):
        database = SlicedDatabaseWindow(mode='binding-energy')
        database.files_chosen.connect(self.load_sliced_files_online)

        self.sub_windows.update({str(id(database)): database})

    def open_photon_energy_sliceddatabase_browser(self):
        database = SlicedDatabaseWindow(mode='photon-energy')
        database.files_chosen.connect(self.load_sliced_file_online)

        self.sub_windows.update({str(id(database)): database})

    def open_cubefile_sliceddatabase_browser(self):
        database = SlicedDatabaseWindow(mode='cubefile')
        database.files_chosen.connect(self.load_cubefile_as_sliced_online)

        self.sub_windows.update({str(id(database)): database})

    def load_sliced_files_online(self, URLs):
        # Load one or more cube files as sliced data[BE, kx, ky]

        log = logging.getLogger('kmap')
        log.info('Loading .cube file(s) as sliced data[BE, kx, ky]...')

        self.tab_widget.open_sliced_data_tab_by_URLs(URLs)

    def load_sliced_file_online(self, URL):
        # Load one cube file as sliced data[photon_energy, kx, ky]

        log = logging.getLogger('kmap')
        log.info('Loading .cube file as sliced data[photon_energy, kx, ky]...')

        self.tab_widget.open_sliced_data_tab_by_URL(URL)

    def load_cubefile_as_sliced_online(self, URL):
        # Load one cube file as sliced psi[x,y,z] or psik[kx,ky,kz]

        log = logging.getLogger('kmap')
        log.info(
            'Loading .cube file as sliced psi[x,y,z] or psik[kx,ky,kz] ...')

        self.tab_widget.open_sliced_data_tab_by_cube(URL)

    def load_cube_files_online(self, URLs):
        if not URLs:
            # No file chosen
            logging.getLogger('kmap').info('No file chosen')
            return

        # Get Tab to load to
        tab = self.tab_widget.get_orbital_tab_to_load_to()

        # Load URL to Tab
        for URL in URLs:
            try:
                tab.add_orbital_from_online(URL[0], URL[1])

            except ValueError:
                logging.getLogger('kmap').error(
                    'URL "%s" is invalid. File could not be loaded.' % URL)

    def load_cube_files_locally(self):
        # Load one or more new cube files

        extensions = 'cube files (*.cube);; All Files (*)'

        path = config.get_key('paths', 'cube_start')
        if path == 'None':
            paths, _ = QFileDialog.getOpenFileNames(
                None, 'Open file(s)', filter=extensions)
        else:
            start_path = str(__directory__ / path)
            paths, _ = QFileDialog.getOpenFileNames(
                None, 'Open file(s)', str(start_path), extensions)

        if not paths:
            # No file chosen
            logging.getLogger('kmap').info('No file chosen')
            return

        # Get Tab to load to
        tab = self.tab_widget.get_orbital_tab_to_load_to()

        # Load Data to Tab
        for path in paths:
            tab.add_orbital_from_filepath(path)

    def open_log_file(self):
        # Open log file

        path = __directory__ / '../default.log'
        title = 'Log File'

        self.tab_widget.open_file_tab(
            path, title, editable=False, richText=False)

    def open_mod_log_file(self):
        # Open module log file

        path = __directory__ / '../modules.log'
        title = 'Modules Log File'

        self.tab_widget.open_file_tab(
            path, title, editable=False, richText=False)

    def open_general_settings(self):
        # Open general user settings

        path = __directory__ / 'config/settings_user.ini'
        title = 'General Settings'

        self.tab_widget.open_file_tab(
            path, title, editable=True, richText=False)

    def open_orbital_data_tab(self):
        self.tab_widget.open_orbital_data_tab()

    def open_profile_tab(self):
        self.tab_widget.open_profile_tab()

    def open_logging_settings(self):
        # Open logging user settings

        path = __directory__ / 'config/logging_user.ini'
        title = 'Logging Settings'

        self.tab_widget.open_file_tab(
            path, title, editable=True, richText=False)

    def open_general_default_settings(self):
        # Open general default settings

        path = __directory__ / 'config/settings_default.ini'
        title = 'General Settings (Default)'

        self.tab_widget.open_file_tab(
            path, title, editable=False, richText=False)

    def open_logging_default_settings(self):
        # Open logging default settings

        path = __directory__ / 'config/logging_default.ini'
        title = 'Logging Settings (Default)'

        self.tab_widget.open_file_tab(
            path, title, editable=False, richText=False)

    def open_readme(self):
        # Open a README page

        path = __directory__ / 'resources/texts/readme.txt'
        title = 'README'

        self.tab_widget.open_file_tab(
            path, title, editable=False, richText=True)

    def open_welcome(self):
        # Open a welcome page

        path = __directory__ / 'resources/texts/welcome.txt'
        title = 'Welcome'

        self.tab_widget.open_file_tab(
            path, title, editable=False, richText=True)

    def open_about(self):
        # Open an about window

        path = __directory__ / 'resources/texts/about.txt'
        title, text = self.model.get_about_text(path)
        QMessageBox.about(self, title, text)

    def open_lmfit_tab(self, tabs):
        if len(tabs) != 2 or tabs[0] is None or tabs[1] is None:
            return

        self.tab_widget.open_lmfit_tab(*tabs)

    def open_tab_choose_window(self):
        sliced_tabs = self.tab_widget.get_tabs_of_type(SlicedDataTab)
        orbital_tabs = self.tab_widget.get_tabs_of_type(OrbitalDataTab)

        if len(sliced_tabs) == 0 or len(orbital_tabs) == 0:
            return

        elif len(sliced_tabs) == 1 and len(orbital_tabs) == 1:
            tabs = [*sliced_tabs, *orbital_tabs]
            if self.sender() == self.open_splitview_tab_action:
                self.open_split_view_tab(tabs)
            else:
                self.open_lmfit_tab(tabs)

        else:
            self.tab_choose_window = TabChooseWindow(sliced_tabs,
                                                     orbital_tabs)
            if self.sender() == self.open_splitview_tab_action:
                self.tab_choose_window.tabs_chosen.connect(
                    self.open_split_view_tab)
            else:
                self.tab_choose_window.tabs_chosen.connect(self.open_lmfit_tab)

    def open_split_view_tab(self, tabs):
        if len(tabs) != 2 or tabs[0] is None or tabs[1] is None:
            return

        self.tab_widget.open_split_view_tab(*tabs)

    def open_in_matplotlib(self):
        current_tab = self.tab_widget.get_current_tab()

        if hasattr(current_tab, 'display_in_matplotlib'):
            window = current_tab.display_in_matplotlib()
            self.sub_windows.update({str(id(window)): window})

    def export_to_txt(self):
        current_tab = self.tab_widget.get_current_tab()

        if hasattr(current_tab, 'export_to_txt'):
            text = current_tab.export_to_txt()

            file_name, _ = QFileDialog.getSaveFileName(self, 'Save .txt File')

            with open(file_name, 'w') as file:
                file.write(text)

    def export_to_hdf5(self):
        current_tab = self.tab_widget.get_current_tab()

        if hasattr(current_tab, 'export_to_hdf5'):
            current_tab.export_to_hdf5()

    def reload_settings(self):
        # Reload the settings
        config.setup()

    def duplicate_tab(self):
        self.tab_widget.duplicate_tab()

    def save_project(self):
        tabs = self.tab_widget.get_all_tabs()
        tab_saves = []
        for i, tab in enumerate(tabs):
            save, dependencies = tab.save_state()
            dependencies = [self.tab_widget.get_index_of(
                tab) for tab in dependencies]
            tab_saves.append([i, [type(tab).__name__, save], dependencies])

        path = config.get_key('paths', 'project_start')
        if path == 'None':
            file_name, _ = QFileDialog.getSaveFileName(
                None, 'Save Project File (*.kmap)')
        else:
            start_path = str(__directory__ / path)
            file_name, _ = QFileDialog.getSaveFileName(
                None, 'Save Project File (*.kmap)', str(start_path))

        if not file_name.endswith('.kmap'):
            file_name += '.kmap'

        # pickle.dump(tab_saves, open(file_name, 'wb'))
        self.compress_pickle_dump(file_name, tab_saves)

    def compress_pickle_dump(self, title, data):
        with bz2.BZ2File(title, 'w') as f:
            pickle.dump(data, f)

    def decompress_pickle(self, file):
        data = bz2.BZ2File(file, 'rb')
        data = pickle.load(data)
        return data

    def load_project(self):
        path = config.get_key('paths', 'project_start')
        if path == 'None':
            file_names, _ = QFileDialog.getOpenFileNames(
                None, 'Load Project File (*.kmap)')
        else:
            start_path = str(__directory__ / path)
            file_names, _ = QFileDialog.getOpenFileNames(
                None, 'Load Project File (*.kmap)', str(start_path))

        if not file_names:
            return

        file_path = file_names[0]
        # save = pickle.load(open(file_path, 'rb'))
        save = self.decompress_pickle(file_path)

        data_tabs = [tab_save for tab_save in
                     save if tab_save[1][0] in ['SlicedDataTab',
                                                'OrbitalDataTab']]
        basic_tabs = [tab_save for tab_save in
                      save if tab_save[1][0] not in ['LMFitPlotTab',
                                                     'LMFitTab',
                                                     'LMFitResultTab',
                                                     'SlicedDataTab',
                                                     'OrbitalDataTab',
                                                     'SplitViewTab']]
        lmfit_tabs = [tab_save for tab_save in
                      save if tab_save[1][0] == 'LMFitTab']
        split_tabs = [tab_save for tab_save in save if tab_save[1][0] ==
                      'SplitViewTab']
        result_tabs = [tab_save for tab_save in
                       save if tab_save[1][0] == 'LMFitResultTab']
        plot_tabs = [tab_save for tab_save in
                     save if tab_save[1][0] == 'LMFitPlotTab']

        # Go through all non lmfit tabs first and replace the save
        # for a tab with the tab itself to be used for tabs that
        # depend on it
        for tab_save in basic_tabs:
            index, tab_save, _ = tab_save
            save[index] = self.tab_widget.open_tab_by_save(
                tab_save)

        ID_maps = []
        for tab_save in data_tabs:
            index, tab_save, _ = tab_save
            tab, ID_map = self.tab_widget.open_tab_by_save(
                tab_save)
            save[index] = tab
            ID_maps.append(ID_map)

        for tab_save in lmfit_tabs:
            index, tab_save, dependencies = tab_save
            sliced_idx, orbital_idx = dependencies
            sliced_tab = save[sliced_idx]
            orbital_tab = save[orbital_idx]

            tab = self.tab_widget.open_tab_by_save(tab_save,
                                                   sliced_tab,
                                                   orbital_tab)

            tab.sliced_data_tab_idx = sliced_idx
            tab.orbital_data_tab_idx = orbital_idx

            save[index] = tab

        for tab_save in split_tabs:
            index, tab_save, dependencies = tab_save
            sliced_idx, orbital_idx = dependencies
            sliced_tab = save[sliced_idx]
            orbital_tab = save[orbital_idx]

            tab = self.tab_widget.open_tab_by_save(tab_save,
                                                   sliced_tab,
                                                   orbital_tab)

            save[index] = tab

        for tab_save in result_tabs:
            index, tab_save, dependencies = tab_save
            lmfit_tab_idx = dependencies[0]
            lmfit_tab = save[lmfit_tab_idx]

            tab = self.tab_widget.open_tab_by_save(
                tab_save, lmfit_tab, ID_maps)

            save[index] = tab

        for tab_save in plot_tabs:
            index, tab_save, dependencies = tab_save
            result_tab_idx = dependencies[0]
            result_tab = save[result_tab_idx]

            self.tab_widget.open_tab_by_save(tab_save, result_tab)

    def closeEvent(self, event):
        for window in self.sub_windows.values():
            try:
                window.close()

            except Exception:
                pass

        event.accept()

    def close_current_tab(self):
        tab = self.tab_widget.get_current_tab()

        if tab is not None:
            tab.close()

    def _setup(self):
        self._set_misc()
        self._initialize_shortcuts()

    def _set_misc(self):
        x = int(config.get_key('app', 'x'))
        y = int(config.get_key('app', 'y'))
        w = int(config.get_key('app', 'w'))
        h = int(config.get_key('app', 'h'))
        self.setGeometry(x, y, h, w)
        if config.get_key('app', 'fullscreen') == 'True':
            self.showMaximized()
        self.setWindowTitle('kMap.py')
        self.setWindowIcon(QIcon(str(__directory__ /
                                     'resources/images/icon.png')))
        self.show()

    def _initialize_shortcuts(self):
        actions = [self.load_hdf5_action,
                   self.load_sliced_from_binding_energy_action,
                   self.load_sliced_from_photon_energy_action,
                   self.show_matplotlib,
                   self.log_file_action, self.load_cube_online_action,
                   self.load_cube_file_action,
                   self.open_lmfit_tab_action,
                   self.settings_action, self.duplicate_tab_action,
                   self.close_tab_action, self.save_project_action,
                   self.load_project_action,
                   self.general_default_action,
                   self.general_action,
                   self.logging_default_action,
                   self.logging_action, self.open_sim_tab_action,
                   self.open_profile_tab_action, self.mod_log_file_action,
                   self.export_txt, self.load_sliced_from_cubefile_action,
                   self.welcome_action, self.about_action, self.readme_action,
                   self.export_hdf5, self.open_splitview_tab_action]

        alias = ['load_hdf5', 'load_sliced_from_binding_energy_action',
                 'load_sliced_from_photon_energy_action',
                 'show_matplotlib', 'open_log',
                 'load_cube_online', 'load_cube_file',
                 'open_lmfit', 'reload_settings',
                 'duplicate_tab_action', 'close_tab_action',
                 'save_project_action', 'load_project_action',
                 'open_general_default', 'open_general_user',
                 'open_logging_default', 'open_logging_user',
                 'open_sim_tab_action', 'open_profile_tab_action',
                 'mod_log_file_action', 'export_txt',
                 'load_sliced_from_cubefile_action',
                 'welcome_action', 'about_action', 'readme_action',
                 'export_hdf5', 'split_view']

        for action, alias in zip(actions, alias):
            shortcut = config.get_key('shortcut', alias, file='shortcut')
            if shortcut != 'None':
                action.setShortcut(QKeySequence(shortcut))

        self.ref_action.setShortcut(QKeySequence('Ctrl+r'))

    def _connect(self):
        # File menu
        self.load_hdf5_action.triggered.connect(self.load_hdf5_files)
        self.load_sliced_from_binding_energy_action.triggered.connect(
            self.open_binding_energy_sliceddatabase_browser)
        self.load_sliced_from_photon_energy_action.triggered.connect(
            self.open_photon_energy_sliceddatabase_browser)
        self.load_sliced_from_cubefile_action.triggered.connect(
            self.open_cubefile_sliceddatabase_browser)
        self.load_cube_file_action.triggered.connect(
            self.load_cube_files_locally)
        self.load_cube_online_action.triggered.connect(
            self.open_orbitaldatabase_browser)
        self.log_file_action.triggered.connect(self.open_log_file)
        self.mod_log_file_action.triggered.connect(self.open_mod_log_file)
        self.show_matplotlib.triggered.connect(self.open_in_matplotlib)
        self.export_txt.triggered.connect(self.export_to_txt)
        self.export_hdf5.triggered.connect(self.export_to_hdf5)
        self.duplicate_tab_action.triggered.connect(self.duplicate_tab)
        self.close_tab_action.triggered.connect(self.close_current_tab)
        self.save_project_action.triggered.connect(self.save_project)
        self.load_project_action.triggered.connect(self.load_project)

        # Tabs menu
        self.open_sim_tab_action.triggered.connect(self.open_orbital_data_tab)
        self.open_profile_tab_action.triggered.connect(self.open_profile_tab)
        self.open_lmfit_tab_action.triggered.connect(
            self.open_tab_choose_window)
        self.open_splitview_tab_action.triggered.connect(
            self.open_tab_choose_window)

        # Preferences menu
        self.general_action.triggered.connect(self.open_general_settings)
        self.logging_action.triggered.connect(self.open_logging_settings)
        self.general_default_action.triggered.connect(
            self.open_general_default_settings)
        self.logging_default_action.triggered.connect(
            self.open_logging_default_settings)
        self.settings_action.triggered.connect(self.reload_settings)

        # Help menu
        self.readme_action.triggered.connect(self.open_readme)
        self.welcome_action.triggered.connect(self.open_welcome)
        self.about_action.triggered.connect(self.open_about)



from itertools import chain
import numpy as np

from pyqtgraph.opengl import (
    GLViewWidget, GLGridItem, GLLinePlotItem,
    MeshData, GLMeshItem, GLSurfacePlotItem)
from pyqtgraph import isosurface

from kmap.controller.pyqtgraphplot import PyQtGraphPlot
from kmap.library.qwidgetsub import FixedSizeWidget
from kmap.library.misc import get_rotation_axes, energy_to_k
from kmap.config.config import config


class MiniKSpacePlot(FixedSizeWidget, PyQtGraphPlot):

    def __init__(self, *args, **kwargs):
        width = 250
        self.ratio = 1
        self.ID = None

        super(MiniKSpacePlot, self).__init__(300, 1, *args, **kwargs)
        self._setup()

    def plot(self, plot_data, ID):
        self.ID = ID

        PyQtGraphPlot.plot(self, plot_data)

    def _setup(self):
        PyQtGraphPlot._setup(self)

        self.ui.histogram.hide()
        self.view.showAxis('bottom', show=False)
        self.view.showAxis('left', show=False)
        self.view.showAxis('top', show=False)
        self.view.showAxis('right', show=False)
        self.view.setAspectLocked(lock=True, ratio=self.ratio)


class MiniRealSpacePlot(GLViewWidget):

    def __init__(self, *args, **kwargs):
        self.options = None
        self.grid = None
        self.bonds = []
        self.photon = None
        self.photon_parameters = ['p', 45, 0]
        self.mesh = []
        self.orbital = None
        self.orientation = [0, 0, 0]

        super(MiniRealSpacePlot, self).__init__(*args, **kwargs)
        self._setup()

        self.show()

    def set_orbital(self, orbital):
        self.orbital = orbital
        self.orientation = [0, 0, 0]

        self.refresh_plot()

    def set_options(self, options):
        self.options = options
        self._connect()

    def rotate_orbital(self, phi=0, theta=0, psi=0):
        old_axes = get_rotation_axes(*self.orientation[:2])
        new_axes = get_rotation_axes(phi, theta)
        for item in chain(self.bonds, self.mesh):
            # Undo current orientation
            self._rotate_item(item, old_axes, *self.orientation,
                              backward=True)
            # Actuallly rotate to new orientation
            self._rotate_item(item, new_axes, phi, theta, psi, backward=False)

        self.orientation = [phi, theta, psi]

    def rotate_photon(self, polarization='p', alpha=0, beta=0):
        if self.photon is not None:
            self.removeItem(self.photon)
            self.photon = None

        self.photon_parameters = [polarization, alpha, beta]
        self._refresh_photon()

    def refresh_plot(self):
        self._refresh_grid()
        self._refresh_bonds()
        self._refresh_photon()
        self._refresh_mesh()

    def reset_camera(self, distance=75, elevation=90, azimuth=-90):
        # View from top
        self.setCameraPosition(distance=distance, elevation=elevation,
                               azimuth=azimuth)

    def toggle_show_grid(self, state):
        if self.grid is not None:
            self.grid.setVisible(state)

    def toggle_show_bonds(self, state):
        if self.bonds:
            for bond in self.bonds:
                bond.setVisible(state)

    def toggle_show_photon(self, state):
        if self.photon is not None:
            self.photon.setVisible(state)

    def toggle_show_mesh(self, state):
        if self.mesh:
            for mesh in self.mesh:
                mesh.setVisible(state)

    def wheelEvent(self, event, *args, **kwargs):
        event.accept()

        super().wheelEvent(event, *args, **kwargs)

    def _refresh_mesh(self):
        if self.mesh:
            for mesh in self.mesh:
                self.removeItem(mesh)

            self.mesh = []

        if self.orbital is None or not self.options.is_show_mesh():
            return

        data = self.orbital.psi['data']
        plus_mesh = self._get_iso_mesh(data, 'red', 1)
        minus_mesh = self._get_iso_mesh(data, 'blue', -1)
        self.addItem(plus_mesh)
        self.addItem(minus_mesh)

        self.mesh = [plus_mesh, minus_mesh]

        # Updating the mesh after iso val change would leave it
        # unrotated
        axes = get_rotation_axes(*self.orientation[:2])
        for item in self.mesh:
            self._rotate_item(item, axes, *self.orientation, backward=False)

    def _refresh_bonds(self):
        if self.bonds:
            for bond in self.bonds:
                self.removeItem(bond)

            self.bonds = []

        if self.orbital is None or not self.options.is_show_bonds():
            return

        color = (0.8, 0.8, 0.8, 1)  # light gray

        for bond in self.orbital.get_bonds():
            new_bond = GLLinePlotItem(pos=bond, color=color,
                                      width=5, antialias=True)
            self.bonds.append(new_bond)

            self.addItem(new_bond)

    def _refresh_photon(self):
        # Couldn't find the bug without second part of if so I removed
        # it for now
        if self.photon:
            self.removeItem(self.photon)
            self.photon = None

        polarization, alpha, beta = self.photon_parameters

        if self.orbital is None or not self.options.is_show_photon():
            return

        color = (1, 1, 0.0, 1)  # color for wavy light ray (yellow)
        ray_length = 50  # length of wavy light ray
        amplitude = 5  # amplitude of oscillation
        wavelength = 5  # wavelength of oscillation
        n_points = 200  # number of points along wavy light ray

        alpha = alpha * np.pi / 180
        beta = (180 + beta) * np.pi / 180
        direction = [np.sin(alpha) * np.cos(beta),
                     np.sin(alpha) * np.sin(beta), np.cos(alpha)]

        x0 = np.linspace(0, ray_length * direction[0], n_points)
        y0 = np.linspace(0, ray_length * direction[1], n_points)
        z0 = np.linspace(0, ray_length * direction[2], n_points)
        t = np.linspace(0, ray_length, n_points)

        if polarization == 's':
            pol_1 = [np.sin(beta), -np.cos(beta), 0]
            pol_2 = [0, 0, 0]

        elif polarization == 'p':
            pol_1 = [0, 0, 0]
            pol_2 = [np.cos(alpha) * np.cos(beta), np.cos(alpha) *
                     np.sin(beta), -np.sin(alpha)]

        # ... to be updated ...
        elif polarization == 'unpolarized':
            pol_1 = [0, 0, 0]
            s_share = float(config.get_key('orbital', 's_share'))
            p_share = 1 - s_share
            pol_2 = [s_share * np.sin(beta) +
                     p_share * np.cos(alpha) * np.cos(beta),
                     -s_share * np.cos(beta) +
                     p_share * np.cos(alpha) * np.sin(beta),
                     -p_share * np.sin(alpha)]

        elif polarization == 'C+':
            pol_1 = [np.sin(beta), -np.cos(beta), 0]
            pol_2 = [np.cos(alpha) * np.cos(beta), np.cos(alpha) *
                     np.sin(beta), -np.sin(alpha)]

        elif polarization == 'C-':
            pol_1 = [np.sin(beta), -np.cos(beta), 0]
            pol_2 = [-np.cos(alpha) * np.cos(beta), -
            np.cos(alpha) * np.sin(beta), +np.sin(alpha)]

        # show C+ spiral ... until I have a better idea ...
        elif polarization == 'CDAD':
            pol_1 = [np.sin(beta), -np.cos(beta), 0]
            pol_2 = [np.cos(alpha) * np.cos(beta), np.cos(alpha) *
                     np.sin(beta), -np.sin(alpha)]

        dx = amplitude * pol_1[0] * np.cos(2 * np.pi * t / wavelength) + \
             amplitude * pol_2[0] * np.sin(2 * np.pi * t / wavelength)
        dy = amplitude * pol_1[1] * np.cos(2 * np.pi * t / wavelength) + \
             amplitude * pol_2[1] * np.sin(2 * np.pi * t / wavelength)
        dz = amplitude * pol_1[2] * np.cos(2 * np.pi * t / wavelength) + \
             amplitude * pol_2[2] * np.sin(2 * np.pi * t / wavelength)

        pos = np.array([x0 + dx, y0 + dy, z0 + dz]).T

        photon_line = GLLinePlotItem(
            pos=pos, color=color, width=5, antialias=True)
        photon_line.setGLOptions('translucent')
        self.photon = photon_line
        self.addItem(photon_line)

    def _refresh_grid(self):
        if self.grid is not None:
            self.removeItem(self.grid)
            self.grid = None

        if self.orbital is None or not self.options.is_show_grid():
            return

        self.grid = GLGridItem()
        dx, dy, dz = [self.orbital.psi[key] for key in ['dx', 'dy', 'dz']]
        self.grid.scale(1 / dx, 1 / dy, 1 / dz)
        self.addItem(self.grid)

    def _rotate_item(self, item, axes, phi, theta, psi, backward=False):
        if backward:
            # Undo Rotation in reverse order
            item.rotate(psi, axes[2][0], axes[2][1], axes[2][2], local=True)
            item.rotate(theta, axes[1][0], axes[1][1], axes[1][2], local=True)
            item.rotate(phi, axes[0][0], axes[0][1], axes[0][2], local=True)

        else:
            item.rotate(-phi, axes[0][0], axes[0][1], axes[0][2], local=True)
            item.rotate(-theta, axes[1][0], axes[1][1], axes[1][2], local=True)
            item.rotate(-psi, axes[2][0], axes[2][1], axes[2][2], local=True)

    def _get_iso_mesh(self, data, color, sign):
        iso_val = sign * self.options.get_iso_val()
        vertices, faces = isosurface(data, iso_val * data.max())
        nx, ny, nz = data.shape

        # Center Isosurface around Origin
        vertices[:, 0] = vertices[:, 0] - nx / 2
        vertices[:, 1] = vertices[:, 1] - ny / 2
        vertices[:, 2] = vertices[:, 2] - nz / 2

        mesh_data = MeshData(vertexes=vertices, faces=faces)
        colors = np.zeros((mesh_data.faceCount(), 4), dtype=float)
        # Sets color to Red (RGB, 0 = red, 1 = green, 2 = blue)
        if color == 'red':
            colors[:, 0] = 1.0
            colors[:, 1] = 0.2
            colors[:, 2] = 0.2

        elif color == 'blue':
            colors[:, 0] = 0.2
            colors[:, 1] = 0.2
            colors[:, 2] = 1.0
        # Transparency (I guess)
        colors[:, 3] = 1
        mesh_data.setFaceColors(colors)

        mesh = GLMeshItem(meshdata=mesh_data, smooth=True,
                          shader='edgeHilight')
        mesh.setGLOptions('translucent')

        return mesh

    def _setup(self):
        # Set Fixed Size. Due to some unknown reason subclassing from a
        # second class like FixedSizeWidget while also
        # subclassing from gl.GLViewWidget throws error
        width = 275
        ratio = 1
        height = width * ratio
        self.resize(width, height)
        self.setMaximumSize(width, height)
        self.setMinimumSize(width, height)

        self.reset_camera()

    def _connect(self):
        self.options.set_camera.connect(self.reset_camera)
        self.options.show_grid_changed.connect(self.toggle_show_grid)
        self.options.show_mesh_changed.connect(self.toggle_show_mesh)
        self.options.show_bonds_changed.connect(self.toggle_show_bonds)
        self.options.show_photon_changed.connect(self.toggle_show_photon)
        self.options.iso_val_changed.connect(self._refresh_mesh)


class Mini3DKSpacePlot(GLViewWidget):

    def __init__(self, *args, **kwargs):
        self.ID = None
        self.options = None
        self.grid = None
        self.hemisphere = None
        self.mesh = None
        self.orbital = None
        self.orientation = [0, 0, 0]
        self.E_kin = 30

        super(Mini3DKSpacePlot, self).__init__(*args, **kwargs)
        self._setup()

        self.show()

    def set_orbital(self, orbital, ID):
        self.ID = ID
        self.orbital = orbital
        self.orientation = [0, 0, 0]

        self.refresh_plot()

    def set_options(self, options):
        self.options = options
        self._connect()

    def rotate_orbital(self, phi=0, theta=0, psi=0):
        old_axes = get_rotation_axes(*self.orientation[:2])
        new_axes = get_rotation_axes(phi, theta)

        # Undo current orientation
        self._rotate_item(self.mesh, old_axes, *self.orientation,
                          backward=True)
        # Actuallly rotate to new orientation
        self._rotate_item(self.mesh, new_axes, phi, theta, psi, backward=False)

        self.orientation = [phi, theta, psi]

    def change_energy(self, E_kin):
        self.E_kin = E_kin

        self._refresh_hemisphere()

    def refresh_plot(self):
        self._refresh_hemisphere()
        self._refresh_grid()
        self._refresh_mesh()

    def reset_camera(self, distance=75, elevation=90, azimuth=-90):
        # View from top
        self.setCameraPosition(distance=distance, elevation=elevation,
                               azimuth=azimuth)

    def toggle_show_grid(self, state):
        if self.grid is not None:
            self.grid.setVisible(state)

    def toggle_show_hemisphere(self, state):
        if self.hemisphere is not None:
            self.hemisphere.setVisible(state)

    def _refresh_mesh(self):
        if self.mesh is not None:
            self.removeItem(self.mesh)

            self.mesh = None

        if self.orbital is None:
            return

        data = self.orbital.psik['data']
        self.mesh = self._get_iso_mesh(data)
        self.addItem(self.mesh)

        # Updating the mesh after iso val change would leave it
        # unrotated
        axes = get_rotation_axes(*self.orientation[:2])
        self._rotate_item(self.mesh, axes, *self.orientation, backward=False)

    def wheelEvent(self, event, *args, **kwargs):
        event.accept()

        super().wheelEvent(event, *args, **kwargs)

    def _refresh_hemisphere(self):
        if self.hemisphere is not None:
            self.removeItem(self.hemisphere)
            self.hemisphere = None

        if self.orbital is None or not self.options.is_show_hemisphere():
            return

        k = energy_to_k(self.E_kin)
        kx, ky, kz = [self.orbital.psik[key] for key in ['kx', 'ky', 'kz']]
        self.dx, self.dy, self.dz = [
            kx[1] - kx[0], ky[1] - ky[0], kz[1] - kz[0]]

        # this produces a hemisphere using the GLSurfacePlotItem, however,
        # the resulting hemisphere appears a little ragged at the circular
        # boundary
        # x = np.linspace(-k / self.dx, k / self.dx, 200)
        # y = np.linspace(-k / self.dy, k / self.dy, 200)
        # X, Y = np.meshgrid(x, y)
        # Z = np.sqrt(k**2 / (self.dx * self.dy) - X**2 - Y**2)

        # self.hemisphere = GLSurfacePlotItem(x=x, y=y, z=Z, color=(
        #    0.5, 0.5, 0.5, 0.7), shader='edgeHilight')
        # self.hemisphere.setGLOptions('translucent')

        # NEW method:
        nz = len(kz)
        X, Y, Z = np.meshgrid(kx / self.dx, ky / self.dy,
                              kz[nz // 2:] / self.dz)
        data = X**2 + Y**2 + Z**2
        iso = k**2 / (self.dx * self.dy)
        color = (0.5, 0.5, 0.5, 0.8)
        self.hemisphere = self._get_iso_mesh(data, iso, color, z_shift=False)

        self.addItem(self.hemisphere)

    def _refresh_grid(self):
        if self.grid is not None:
            self.removeItem(self.grid)
            self.grid = None

        if self.orbital is None or not self.options.is_show_grid():
            return

        self.grid = GLGridItem()
        self.grid.setSize(x=8, y=8, z=8)
        self.grid.scale(1 / self.dx, 1 / self.dy, 1 / self.dz)
        self.addItem(self.grid)

    def _rotate_item(self, item, axes, phi, theta, psi, backward=False):
        if backward:
            # Undo Rotation in reverse order
            item.rotate(psi, axes[2][0], axes[2][1], axes[2][2], local=True)
            item.rotate(theta, axes[1][0], axes[1][1], axes[1][2], local=True)
            item.rotate(phi, axes[0][0], axes[0][1], axes[0][2], local=True)

        else:
            item.rotate(-phi, axes[0][0], axes[0][1], axes[0][2], local=True)
            item.rotate(-theta, axes[1][0], axes[1][1], axes[1][2], local=True)
            item.rotate(-psi, axes[2][0], axes[2][1], axes[2][2], local=True)

    def _get_iso_mesh(self, data, iso=None, color=(1, 0.2, 0.2, 1),
                      z_shift=True):
        if iso is None:
            iso_val = self.options.get_iso_val() * data.max()
        else:
            iso_val = iso

        vertices, faces = isosurface(data, iso_val)
        nx, ny, nz = data.shape

        # Center Isosurface around Origin
        vertices[:, 0] = vertices[:, 0] - nx / 2
        vertices[:, 1] = vertices[:, 1] - ny / 2
        if z_shift:
            vertices[:, 2] = vertices[:, 2] - nz / 2

        mesh_data = MeshData(vertexes=vertices, faces=faces)
        colors = np.zeros((mesh_data.faceCount(), 4), dtype=float)
        for idx in range(4):
            colors[:, idx] = color[idx]

        mesh_data.setFaceColors(colors)

        mesh = GLMeshItem(meshdata=mesh_data, smooth=True,
                          shader='edgeHilight')
        mesh.setGLOptions('translucent')

        return mesh

    def _setup(self):
        # Set Fixed Size. Due to some unknown reason subclassing from a
        # second class like FixedSizeWidget while also
        # subclassing from gl.GLViewWidget throws error
        width = 275
        ratio = 1
        height = width * ratio
        self.resize(width, height)
        self.setMaximumSize(width, height)
        self.setMinimumSize(width, height)

        self.reset_camera()

    def _connect(self):
        self.options.set_camera.connect(self.reset_camera)
        self.options.show_grid_changed.connect(self.toggle_show_grid)
        self.options.show_hemisphere_changed.connect(
            self.toggle_show_hemisphere)
        self.options.iso_val_changed.connect(self._refresh_mesh)
